@@ -19,46 +19,7 @@
 #pragma interface			/* gcc class implementation */
 #endif
 
-#include "my_sys.h"                    /* alloc_root, TRASH, MY_WME,
-                                          MY_FAE, MY_ALLOW_ZERO_PTR */
-#include "m_string.h"                           /* bfill */
-
-THD *thd_get_current_thd();
-
-/* mysql standard class memory allocator */
-
-class Sql_alloc
-{
-public:
-  static void *operator new(size_t size) throw ()
-  {
-    return thd_alloc(thd_get_current_thd(), size);
-  }
-  static void *operator new[](size_t size) throw ()
-  {
-    return thd_alloc(thd_get_current_thd(), size);
-  }
-  static void *operator new[](size_t size, MEM_ROOT *mem_root) throw ()
-  { return alloc_root(mem_root, size); }
-  static void *operator new(size_t size, MEM_ROOT *mem_root) throw ()
-  { return alloc_root(mem_root, size); }
-  static void operator delete(void *ptr, size_t size) { TRASH_FREE(ptr, size); }
-  static void operator delete(void *ptr, MEM_ROOT *mem_root)
-  { /* never called */ }
-  static void operator delete[](void *ptr, MEM_ROOT *mem_root)
-  { /* never called */ }
-  static void operator delete[](void *ptr, size_t size) { TRASH_FREE(ptr, size); }
-#ifdef HAVE_valgrind
-  bool dummy_for_valgrind;
-  inline Sql_alloc() :dummy_for_valgrind(0) {}
-  inline ~Sql_alloc() {}
-#else
-  inline Sql_alloc() {}
-  inline ~Sql_alloc() {}
-#endif
-
-};
-
+#include "sql_alloc.h"
 
 /**
   Simple intrusive linked list.
@@ -67,6 +28,7 @@ public:
           a pointer to the first element in the list and a indirect
           reference to the last element.
 */
+
 template <typename T>
 class SQL_I_List :public Sql_alloc
 {
@@ -176,6 +138,13 @@ public:
       first == rhs.first &&
       last == rhs.last;
   }
+  base_list& operator=(const base_list &rhs)
+  {
+    elements= rhs.elements;
+    first= rhs.first;
+    last= elements ? rhs.last : &first;
+    return *this;
+  }
 
   inline void empty() { elements=0; first= &end_of_list; last=&first;}
   inline base_list() { empty(); }
@@ -190,9 +159,7 @@ public:
   */
   inline base_list(const base_list &tmp) :Sql_alloc()
   {
-    elements= tmp.elements;
-    first= tmp.first;
-    last= elements ? tmp.last : &first;
+    *this= tmp;
   }
   /**
     Construct a deep copy of the argument in memory root mem_root.
@@ -202,7 +169,7 @@ public:
   */
   bool copy(const base_list *rhs, MEM_ROOT *mem_root);
   base_list(const base_list &rhs, MEM_ROOT *mem_root) { copy(&rhs, mem_root); }
-  inline base_list(bool error) { }
+  inline base_list(bool) {}
   inline bool push_back(void *info)
   {
     if (((*last)=new list_node(info, &end_of_list)))
@@ -338,16 +305,15 @@ public:
   friend class error_list;
   friend class error_list_iterator;
 
-#ifndef DBUG_OFF
   /*
-    Debugging help: return N-th element in the list, or NULL if the list has
+    Return N-th element in the list, or NULL if the list has
     less than N elements.
   */
-  void *elem(int n)
+  void *elem(uint n)
   {
     list_node *node= first;
     void *data= NULL;
-    for (int i=0; i <= n; i++)
+    for (uint i= 0; i <= n; i++)
     {
       if (node == &end_of_list)
       {
@@ -359,7 +325,6 @@ public:
     }
     return data;
   }
-#endif
 
 #ifdef LIST_EXTRA_DEBUG
   /*
@@ -528,10 +493,10 @@ public:
     base_list(tmp, mem_root) {}
   inline bool push_back(T *a) { return base_list::push_back(a); }
   inline bool push_back(T *a, MEM_ROOT *mem_root)
-  { return base_list::push_back(a, mem_root); }
+  { return base_list::push_back((void*) a, mem_root); }
   inline bool push_front(T *a) { return base_list::push_front(a); }
   inline bool push_front(T *a, MEM_ROOT *mem_root)
-  { return base_list::push_front(a, mem_root); }
+  { return base_list::push_front((void*) a, mem_root); }
   inline T* head() {return (T*) base_list::head(); }
   inline T** head_ref() {return (T**) base_list::head_ref(); }
   inline T* pop()  {return (T*) base_list::pop(); }
@@ -552,9 +517,7 @@ public:
     }
     empty();
   }
-#ifndef DBUG_OFF
-  T *elem(int n) { return (T*)base_list::elem(n); }
-#endif
+  T *elem(uint n) { return (T*) base_list::elem(n); }
 };
 
 
@@ -578,10 +541,10 @@ public:
 template <class T> class List_iterator_fast :public base_list_iterator
 {
 protected:
-  inline T *replace(T *a)   { return (T*) 0; }
-  inline T *replace(List<T> &a) { return (T*) 0; }
-  inline void remove(void)  { }
-  inline void after(T *a)   { }
+  inline T *replace(T *)   { return (T*) 0; }
+  inline T *replace(List<T> &) { return (T*) 0; }
+  inline void remove(void)  {}
+  inline void after(T *)    {}
   inline T** ref(void)	    { return (T**) 0; }
 
 public:
@@ -650,7 +613,7 @@ struct ilink
   {
     return (void*)my_malloc((uint)size, MYF(MY_WME | MY_FAE | ME_FATALERROR));
   }
-  static void operator delete(void* ptr_arg, size_t size)
+  static void operator delete(void* ptr_arg, size_t)
   {
      my_free(ptr_arg);
   }
@@ -670,6 +633,10 @@ struct ilink
   {
     DBUG_ASSERT(prev != 0 && next != 0);
   }
+  inline void assert_not_linked()
+  {
+    DBUG_ASSERT(prev == 0 && next == 0);
+  }
   virtual ~ilink() { unlink(); }		/*lint -e1740 */
 };
 
